// This header must be visible to all MySQL code that:
// 1) creates threads, enters mutexes and uses other inter-thread synchronization
// 2) uses GOV_LOG() to write '/var/log/dbgovernor-mysqld.log'.
// Usually it's #include'd into 'my_thread.h'.
// Requires 'global.c.inc' in the project.

// ACHTUNG: Use ONLY 'GOV_...' MACROS in С code patches!

#pragma once

// --------------------------------------------------------------
// Logging to '/var/log/dbgovernor-mysqld.log' from the patch code.
// Part of so-called "extended logging".
//
#include <governor-mysql/log_defs.h>

extern t_write_log_ex *libgovernor_write_log_ex;
extern t_get_log_ex_properties *libgovernor_get_log_ex_properties;

extern unsigned govlog_enabled_tags;
extern int govlog_unconditional_log_if;	// if true, GOV_LOG_IF(cond, ...) works as if "cond" is true; used to debug logging itself

// Usage: GOV_LOG(L_ERR, "blabla %d", 235)
#define GOV_LOG(tags, fmt, ...) do { \
		if ((tags) & govlog_enabled_tags) \
		{ \
			const char *failure = NULL; \
			if (!libgovernor_write_log_ex) /* libgovernor.so not found or old */ \
				failure = "Governor logging unavailable"; \
			else if (libgovernor_write_log_ex(tags, 1, __FILE__, __LINE__, __FUNCTION__, fmt, ##__VA_ARGS__)) /* governor not connected, or "/var/log/dbgovernor-mysqld.log" failed to open, or internal error */ \
				failure = "Governor logging failed"; \
			if (failure && ((tags) & _L_MASK_ALWAYS_ENABLED)) /* failed, but important, let's print by other means */ \
			{ \
				if ((tags) & _L_MASK_ERRORS) \
					sql_print_error(		"(%s) " fmt, failure, ##__VA_ARGS__); \
				else \
					sql_print_information(	"(%s) " fmt, failure, ##__VA_ARGS__); \
			} \
		} \
	} while(0)

#define GOV_LOG_IF(cond, tags, fmt, ...) do { \
	if ((cond) || govlog_unconditional_log_if) \
		GOV_LOG(tags, fmt, ##__VA_ARGS__); \
	} while (0)



// --------------------------------------------------------------
// Interception of MySQL synchronization primitives to integrate them with LVE

// LVE slot reservation.
extern void (*govlve_slot_reserve)();
extern void (*govlve_slot_release)();
// Use these macros where 'GOV_LVE_MUTEX_...()' and 'GOV_LVE_...THREAD_CREATE()' are inapplicable:
#define GOV_LVE_SLOT_RESERVE()				govlve_slot_reserve()
#define GOV_LVE_SLOT_RELEASE()				govlve_slot_release()

// pthread_mutex_...() substitution:
extern int (*govlve_mutex_lock)   (pthread_mutex_t *mutex);
extern int (*govlve_mutex_trylock)(pthread_mutex_t *mutex);
extern int (*govlve_mutex_unlock) (pthread_mutex_t *mutex);
#define GOV_LVE_MUTEX_LOCK(mutex)			govlve_mutex_lock(mutex)
#define GOV_LVE_MUTEX_TRYLOCK(mutex)		govlve_mutex_trylock(mutex)
#define GOV_LVE_MUTEX_UNLOCK(mutex)			govlve_mutex_unlock(mutex)

// pthread_create() substitution.
// In many patches these helpers are not required,
// because the wrapper around pthread_create() is already defined in the upstream code,
// and we simply add GOV_LVE_SLOT_...() to it.
#define GOV_LVE_PTHREAD_CREATE_DECL() \
	int govlve_pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*routine)(void*), void *arg)
#define GOV_LVE_PTHREAD_CREATE_IMPL() \
	int govlve_pthread_create(pthread_t *__restrict thread, const pthread_attr_t *__restrict attr, void *(*routine)(void*), void *__restrict arg) \
	{ \
		govlve_slot_reserve(); \
		int rc = pthread_create(thread, attr, routine, arg); \
		govlve_slot_release(); \
		return rc; \
	}
#define GOV_LVE_PTHREAD_CREATE(thread, attr, routine, arg)	govlve_pthread_create(thread, attr, routine, arg)
