
// Implementation for 'main.h'.
// Must be included in every project using that header, usually at the bottom of 'mysqld.cc'.

#if !defined(GOV_CFG_SUPPORT_FULL_LVE) && !defined(GOV_CFG_SUPPORT_LIGHT_LVE)
	#error at least one of GOV_CFG_SUPPORT_FULL_LVE and GOV_CFG_SUPPORT_LIGHT_LVE macros should be defined
#endif

#include <dlfcn.h>

static void *libgovernor_library_handle = NULL;

// used only in this file - for Governor init/deinit
static int (*libgovernor_connect_to_server)() = NULL;
static int (*libgovernor_close_sock)() = NULL;

// used outside of this file
int (*libgovernor_send_info_begin)(const char*) = NULL;
int (*libgovernor_send_info_end)(const char*) = NULL;

static void set_libgovernor_main_funcs_null()
{
	libgovernor_connect_to_server = NULL;
	libgovernor_close_sock = NULL;
	libgovernor_send_info_begin = NULL;
	libgovernor_send_info_end = NULL;
	libgovernor_write_log_ex = NULL;
	libgovernor_get_log_ex_properties = NULL;
}

#ifdef __cplusplus
extern "C" {
#endif

static void * (*libgovernor_load_lve_library)() = NULL;			// load library by dl and initialize all pointers to fuctions
static int (*libgovernor_init_lve)() = NULL;					// init_lve wrapper
static void (*libgovernor_destroy_lve)() = NULL;				// destroy_lve wrapper
static int (*libgovernor_enter_lve)(uint32_t *, const char *) = NULL;	// lve_enter wrapper
static int (*libgovernor_reenter_lve)(uint32_t *) = NULL;		// lve_enter with stored pcookie and UID in thread storage
static void (*libgovernor_lve_exit)(uint32_t *) = NULL;			// lve_leave wrapper

#ifdef __cplusplus
}
#endif

static void set_libgovernor_lve_funcs_null()
{
	libgovernor_load_lve_library = NULL;
	libgovernor_init_lve = NULL;
	libgovernor_destroy_lve = NULL;
	libgovernor_enter_lve = NULL;
	libgovernor_reenter_lve = NULL;
	libgovernor_lve_exit = NULL;
}


volatile int gov_regime = govregime_OFF;

void govsql_enable()
{
	gov_regime = govregime_COLLECT_STATS_ONLY;
}

void govsql_enable_reconn()
{
	gov_regime = govregime_COLLECT_STATS_ONLY;
	if (libgovernor_close_sock)
		libgovernor_close_sock();
	if (libgovernor_connect_to_server)
	{
		if (!libgovernor_connect_to_server())
			sql_print_information("Governor reconnected");
		else
			sql_print_error("Governor not reconnected. Failed connection");
	}
}

void govsql_enable_lve()
{
	gov_regime = govregime_COLLECT_STATS_AND_APPLY_LVE;
}

void govsql_enable_reconn_lve()
{
	govsql_enable_reconn();
	gov_regime = govregime_COLLECT_STATS_AND_APPLY_LVE;
}

//Thread dependent variable for thread cookie storage needs for libgovernor_enter_lve, libgovernor_lve_exit
static __thread uint32_t lve_cookie = 0;

int (*govlve_enter)(const char*) = NULL;
void (*govlve_exit)() = NULL;

#if defined(GOV_CFG_SUPPORT_LIGHT_LVE) && defined(GOV_CFG_SUPPORT_FULL_LVE)
	static bool full_mode = false;
#elif defined(GOV_CFG_SUPPORT_LIGHT_LVE)
	#define full_mode false
#elif defined(GOV_CFG_SUPPORT_FULL_LVE)
	#define full_mode true
#else
	#error cannot get here, ensured earlier
#endif

// Thread state - the main thing distinguishing Full mode from the Light one.
// Both in Full and Light modes we keep track of it to detect abnormal situations.
// In Full Mode, we also temporarily exit/re-enter LVE based on it - that is the purpose of Full Mode an its only difference from Light.

// Let's use as short names as possible for values- it's safe enough when we're #include'd at the bottom.
// Unfortunately, we can't use enum with bare 'ON' and 'OFF': 'OFF' is a enum constant in MySQL 5.7.
#define OFF		0	// default thread state, before entering LVE
#define ON		1	// after entering LVE
// values > 2 mean "off in nested mutexes":
//	ON+1				= 2		// after locking 1st mutex (outer) and temporarily exiting LVE
//	ON+2				= 3		// after locking 2nd mutex (nested)
//	ON+3				= 4		// after locking 3rd mutex (nested)
//						...
static __thread int lve = OFF;

// Transition scheme:
// [thread start]  ==>  OFF  ==>  [enter LVE]  ==>  ON  ==>  [exit  LVE temporarily]==>[  lock 1st mutex]  ==>  ON+1  ==>  [  lock 2nd mutex]  ==>  ON+2  ==>  ...
// [thread end]    <==  OFF  <==  [exit  LVE]  <==  ON  <==  [enter LVE back]       <==[unlock 1st mutex]  <==  ON+1  <==  [unlock 2nd mutex]  <==  ON+2  <==  ...
//               ^ ^ ^ ^ ^ ^                       ^^^        ^ ^ ^ ^ ^...
//              (outside LVE)                    (in LVE)    (mutex pit > > >)

// We should be in LVE if and only if lve==ON. If it's >ON, it's essentially !=ON. It's "mutex pit".

#define DBG_PARITY	// Make consistency checks - mostly, parity of calls to the below API. Always defined, used only to isolate these checks in the code.

#ifdef DBG_PARITY
static __thread int dbg_slotdepth = 0;		// ++ on slot_reserve(), -- on slot_release();
#endif // DBG_PARITY

// enter() / exit() are the outermost brackets controlling LVE application.
//
// The following behaviour is assumed, or implemented, or checked - here and below:
// 1) Nesting enter()/exit() pairs looks impossible in the current patches, and is never planned in the future.
// 2) All other primitives - mutex and slot reservation - when started inside enter()/exit() pair,
//    complete (have their pair) within it - i.e., are truly nested.
// 3) slot reservation      outside enter()/exit(): no-op.
// 4) mutex (our additions) outside enter()/exit(): no-op.
// 5) Consequently, if things go right, "lve" and "dbg_slotdepth" have not only the same, but zero values before enter() and after exit().

static int impl__enter(const char *user)
{
	GOV_LOG(L_LVE, "enter LVE, username='%s', lve=%d", user, lve);
#ifdef DBG_PARITY
	// require: LVE off, no slot reservation, no full mode suspension
	GOV_LOG_IF(lve != OFF || dbg_slotdepth, L_ERRSENTRY|L_LVE, "unpaired or misplaced GOV_LVE_... calls detected on entering LVE: username='%s', lve=%d, dbg_slotdepth=%d", user, lve, dbg_slotdepth);
#endif // DBG_PARITY
	lve = OFF;
	int ret = libgovernor_enter_lve(&lve_cookie, user);
	if (ret)
		return ret;
	lve = ON;
	return 0;
}

static void impl__exit()
{
	GOV_LOG(L_LVE, "exit LVE, lve=%d", lve);
#ifdef DBG_PARITY
	// require: LVE on, no slot reservation, no full mode suspension
	GOV_LOG_IF(lve != ON || dbg_slotdepth, L_ERRSENTRY|L_LVE, "unpaired or misplaced GOV_LVE_... calls detected on exiting LVE: lve=%d, dbg_slotdepth=%d", lve, dbg_slotdepth);
#endif // DBG_PARITY
	if (lve == ON || !full_mode)	// in Full Mode - ONLY if things are clear, in Light Mode - in ANY case
		libgovernor_lve_exit(&lve_cookie);
	lve = OFF;		// in ANY case
}

#ifndef GOV_CFG_DONT_INTERCEPT_MUTEX	// In MariaDB, mutex interception was not working for long. It was a mistake, but now, for compatibility, we legitimize this behaviour and make it a configuration option.

#undef pthread_mutex_lock
#undef pthread_mutex_trylock
#undef pthread_mutex_unlock

// The below 3 functions are our substitutes for 'pthread_mutex_...()' API family.
// Inside them, apart from calling the original API, we maintain 'lve' value.
// We do nothing when we're outside LVE (lve==OFF). So, if we're OFF initially, these functions won't ever change it.
// But if we became ON, we start to increment 'lve' on each mutex lock and decrement on each unlock.
// This way, when we transit from ON to ON+1, we know that this increment is caused by the outer (or 'first in the stack') mutex.
// At this moment, we leave LVE for the whole duration of outer mutex locked state - to avoid being LVE-frozen while holding global resources.
// Increment from ON+1 to ON+2, and so on, mean locking nested mutexes (i.e., locking more before unlocking the initial one). In these cases, we do nothing.
// Decrement from ON+2 to ON+1 also means unlocking nested ones.
// But when we decrement from ON+1 to ON, this means that this unlock is done on the outer mutex. And we re-enter LVE.
//
// Our solution works only under Assumption #1 - that the code covered with LVE always locks and releases mutexes in LIFO order.
// If the LVE-protected 3'rd party code:
// 1) is a well-defined piece of business logic
// 2) free of bugs
// 3) uses only trivial IPC schemes,
// then most probably our assumption holds.
// For now, we LVE-wrap only a single MySQL/MariaDB call, and it looks OK.
//
// But the efficiency of our solution depends also on Assumption #2 -
// that the time spent inside those mutexes is significant.
// If the mutex protects a very fast thing, like a reference counter increment, our overhead can be intolerable.

static int impl__mutex_lock(pthread_mutex_t *mutex)
{
	bool exitAndReEnter = false;
	if (lve >= ON)		// in LVE or in mutex pit
	{
		exitAndReEnter = lve == ON && full_mode;	// in LVE and Full Mode
		if (exitAndReEnter)
			libgovernor_lve_exit(&lve_cookie);		// temporarily exit, at least for the duration of locking
	}
	int ret = pthread_mutex_lock(mutex);
	GOV_LOG(L_LVE, "mutex lock, ret=%d, lve=%d", ret, lve);
	if (lve >= ON)		// in LVE or in mutex pit
	{
		if (!ret)	// lock succeeded
			lve++;		// step deeper into mutex pit.
		else		// lock failed
			if (exitAndReEnter && libgovernor_reenter_lve(&lve_cookie))		// need to re-enter, failed to do so
			{
				lve = OFF;		// prematurely finish current LVE invocation for the thread
				GOV_LOG(L_ERR|L_LVE, "failed to re-enter after unsuccessful mutex lock attempt");
			}
	}
	return ret;
}

static int impl__mutex_unlock(pthread_mutex_t *mutex)
{
	int ret = pthread_mutex_unlock(mutex);
	GOV_LOG(L_LVE, "mutex unlock, ret=%d, lve=%d", ret, lve);
	if (!ret && lve > ON)	// successfully unlocked a mutex, while being in mutex pit
	{
		lve--;								// step up from mutex pit,
		bool reEnter = lve == ON && full_mode;	// should be in LVE, and Full Mode
		if (reEnter && libgovernor_reenter_lve(&lve_cookie))		// need to re-enter, failed to do so
		{
			lve = OFF;		// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after mutex unlock");
		}
	}
	return ret;
}

static int impl__mutex_trylock(pthread_mutex_t *mutex)
{
	bool exitAndReEnter = false;
	if (lve >= ON)		// in LVE or in mutex pit
	{
		exitAndReEnter = lve == ON && full_mode;	// in LVE and Full Mode
		if (exitAndReEnter)							// in LVE
			libgovernor_lve_exit(&lve_cookie);			// temporarily exit, at least for the duration of locking
	}
	int ret = pthread_mutex_trylock(mutex);
	GOV_LOG(L_LVE, "mutex trylock, ret=%d, lve=%d", ret, lve);
	if (lve >= ON)		// in LVE or in mutex pit
	{
		if (!ret)	// lock succeeded
			lve++;		// step deeper into mutex pit
		else		// lock failed
			if (exitAndReEnter && libgovernor_reenter_lve(&lve_cookie))		// need to re-enter, failed to do so
			{
				lve = OFF;		// prematurely finish current LVE invocation for the thread
				GOV_LOG(L_ERR|L_LVE, "failed to re-enter after unsuccessful mutex trylock");	// Sentry reporting is done inside libgovernor_reenter_lve()
			}
	}
	return ret;
}

// Prevent 'pthread_mutex_...' reuse, because we #undef'ed them above, and the surrounding code can rely on its own #define's.
// Anyways we must include this whole file at the bottom of some *.c, but just in case:
#define pthread_mutex_lock(m)		error: pthread_mutex_...() spoiled; include this file only at the bottom
#define pthread_mutex_trylock(m)	error: pthread_mutex_...() spoiled; include this file only at the bottom
#define pthread_mutex_unlock(m)		error: pthread_mutex_...() spoiled; include this file only at the bottom

#endif // GOV_CFG_DONT_INTERCEPT_MUTEX

// Slot reservation simulates behaviour of mutex lock/unlock - temporarily exits and then re-enters LVE.
// Compared to mutex interception, these calls are under much greater our control -
// the code where we place them in patches is not as generic and widely used as mutexes.
// Currently the patches look like reserve/release pairs are never nested.
// But:
// 1) I'm not 100% sure,
// 2) they were nestable historically,
// 2) we can add more of them in the future.
// So, we support nesting for them, not just on/off flag.

// Effect is the same as of successful mutex lock.
static void impl__slot_reserve()
{
	GOV_LOG(L_LVE, "reserve slot, lve=%d", lve);
	if (lve >= ON)
	{
		bool exit_ = lve == ON && full_mode;	// in LVE and Full Mode
		if (exit_)
			libgovernor_lve_exit(&lve_cookie);
		lve++;
#ifdef DBG_PARITY
		dbg_slotdepth++;	// although there's no evidence of nested reserve/release pairs potentially can be nested.
#endif // DBG_PARITY
	}
}

// Effect is the same as of successful mutex unlock.
static void impl__slot_release()
{
	GOV_LOG(L_LVE, "release slot, lve=%d", lve);
	if (lve > ON)
	{
#ifdef DBG_PARITY
		dbg_slotdepth--;
		GOV_LOG_IF(dbg_slotdepth < 0, L_ERRSENTRY|L_LVE, "unpaired or misplaced GOV_LVE_... calls detected on unlocking mutex: lve=%d, dbg_slotdepth=%d", lve, dbg_slotdepth);
#endif // DBG_PARITY
		lve--;
		bool reEnter = lve == ON && full_mode;	// in LVE and Full Mode
		if (reEnter && libgovernor_reenter_lve(&lve_cookie))
		{
			lve = OFF;			// prematurely finish current LVE invocation for the thread
			GOV_LOG(L_ERR|L_LVE, "failed to re-enter after slot release");	// Sentry reporting is done inside libgovernor_reenter_lve()
		}
	}
}

void gov_init()
{
	gov_regime = govregime_OFF;

	set_libgovernor_main_funcs_null();
	set_libgovernor_lve_funcs_null();

	const char *path = "libgovernor.so";
	libgovernor_library_handle = dlopen(path, RTLD_LAZY);
	sql_print_information("%s %s", path, libgovernor_library_handle ? "found" : "not found");

	#define LOAD_LIBGOVERNOR_FUNC(fptr, ftype, fname)\
		fptr = NULL;\
		if (!load_error)\
		{\
			fptr = (ftype)dlsym(libgovernor_library_handle, fname);\
			load_error = dlerror();\
			if (load_error)\
				sql_print_warning("Failed to find Governor function '%s': %s", fname, load_error);\
		}

	if (libgovernor_library_handle)
	{
		const char *load_error = 0;

		// It's IMPORTANT that function pointer variables here have DIFFERENT names from the corresponding functions in libgovernor.so.
		// Because dlopen(..., RTLD_LAZY) hides pitfalls. Imagine:
		// 1) 'send_info_begin()' is defined in 'libgovernor.so', and is called somewhere inside it;
		// 2) 'mysqld' executable has a variable 'send_info_begin';
		// 3) 'mysqld' calls dlopen("libgovernor.so", RTLD_LAZY), and everything is alright UNTIL
		// 4) a call to 'send_info_begin()' is done from inside of 'libgovernor.so',
		//    postponed symbol resolution resolves 'send_info_begin()' function address in 'libgovernor.so'
		//    to the address of 'send_info_begin' variable in 'mysqld', and the call crashes.
		//
		// I believe the problem exists with ANY external symbol clash between the exe and .so,
		// but it becomes EXTREMELY dangerous with RTLD_LAZY.

		LOAD_LIBGOVERNOR_FUNC(libgovernor_connect_to_server,	int (*)(),				"connect_to_server");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_send_info_begin,		int (*)(const char*),	"send_info_begin");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_send_info_end,		int (*)(const char*),	"send_info_end");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_close_sock,			int (*)(),				"close_sock");

		if (load_error)
		{
			set_libgovernor_main_funcs_null();
			sql_print_warning("Governor functionality disabled");
		} else
		{
			sql_print_information("All main Governor functions found");

			LOAD_LIBGOVERNOR_FUNC(libgovernor_write_log_ex, t_write_log_ex*, "write_log_ex");
			sql_print_information("Governor logging %s", libgovernor_write_log_ex ? "available" : "unavailable");

			LOAD_LIBGOVERNOR_FUNC(libgovernor_get_log_ex_properties, t_get_log_ex_properties*, "get_log_ex_properties");
			if (libgovernor_get_log_ex_properties)
			{
				libgovernor_get_log_ex_properties(&govlog_enabled_tags, &govlog_unconditional_log_if);
				govlog_enabled_tags |= _L_MASK_ALWAYS_ENABLED;	// ensure unmaskable tags enabled, just in case
			}
		}
	}

	if (libgovernor_connect_to_server)
	{
		if (!libgovernor_connect_to_server())
			sql_print_information("Governor connected");
		else
			sql_print_error("Governor not connected");
	}

	if (libgovernor_library_handle)
	{
		const char *load_error = 0;

		LOAD_LIBGOVERNOR_FUNC(libgovernor_load_lve_library,	void * (*)(),						"governor_load_lve_library");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_init_lve,			int (*)(),							"governor_init_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_destroy_lve,		void (*)(),							"governor_destroy_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_enter_lve,		int (*)(uint32_t *, const char *),	"governor_enter_lve");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_reenter_lve,		int (*)(uint32_t *),				"governor_enter_lve_light");
		LOAD_LIBGOVERNOR_FUNC(libgovernor_lve_exit,			void (*)(uint32_t *),				"governor_lve_exit");

		if (load_error)
		{
			set_libgovernor_lve_funcs_null();
			sql_print_warning("Governor LVE functionality disabled");
		} else
		{
			sql_print_information("All Governor LVE functions found");
		}
	}

	#undef LOAD_LIBGOVERNOR_FUNC

	if (libgovernor_load_lve_library)
	{
		if (!libgovernor_load_lve_library())	// this one returns 0 on _failure_
		{
			sql_print_information("Can't get LVE functions");
			set_libgovernor_lve_funcs_null();
		} else if (libgovernor_init_lve())
		{
			sql_print_error("Governor LVE initialization error");
			set_libgovernor_lve_funcs_null();
		}
	}

	const char *lveMode = "";
	if (!libgovernor_load_lve_library)
		lveMode = "no";
	else
	{
#if defined(GOV_CFG_SUPPORT_LIGHT_LVE) && defined(GOV_CFG_SUPPORT_FULL_LVE)
		const char *lightFile = "/usr/share/lve/dbgovernor/cl_light_lve";
		struct stat st;
		bool light = stat(lightFile, &st) >= 0;
		sql_print_information("'%s' %s", lightFile, light ? "found" : "not found");
		full_mode = !light;
#endif // defined(GOV_CFG_SUPPORT_LIGHT_LVE) && defined(GOV_CFG_SUPPORT_FULL_LVE)
		lveMode = full_mode ? "full" : "light";
		govlve_enter = impl__enter;
		govlve_exit  = impl__exit;
		govlve_slot_reserve = impl__slot_reserve;
		govlve_slot_release = impl__slot_release;
#ifndef GOV_CFG_DONT_INTERCEPT_MUTEX
		govlve_mutex_lock    = impl__mutex_lock;
		govlve_mutex_trylock = impl__mutex_trylock;
		govlve_mutex_unlock  = impl__mutex_unlock;
#endif // GOV_CFG_DONT_INTERCEPT_MUTEX
	}
	sql_print_information("Using %s LVE functionality", lveMode);
	GOV_LOG(L_LIFE, "LVE mode: %s", lveMode);
}

void gov_deinit()
{
	if (libgovernor_destroy_lve)
		libgovernor_destroy_lve();

	if (libgovernor_close_sock)
		libgovernor_close_sock();

	if (libgovernor_library_handle)
	{
		dlclose(libgovernor_library_handle);	// calls libgovernor.so->fini_libgovernor()->close_log(), so after this GOV_LOG() is not fully functional
		libgovernor_library_handle = NULL;	// JIC
	}

	// No libgovernor funcs should be called since now, and we are confident enough - we're right before mysqld process exit.
	// But JIC it's always safer to crash on NULL pointers than on garbage:
	set_libgovernor_main_funcs_null();
	set_libgovernor_lve_funcs_null();
}

