
// This header must be visible to mysqld code that:
// 1) performs startup and shutdown
// 2) enters/exits LVE - e.g., when executing SQL queries.
// Usually it's #include'd into 'mysqld.h'.
// Requires 'main.c.inc' in the project.

// ACHTUNG: Use ONLY 'GOV_...' MACROS in С code patches!

#pragma once

// --------------------------------------------------------------
// Governor initialization and de-initialization:
// loading functions from libgovernor.so, connecting to db_governor, selecting LVE full/light mode, etc.
//
void gov_init();
void gov_deinit();

#define GOV_INIT()		gov_init()
#define GOV_DEINIT()	gov_deinit()


// --------------------------------------------------------------
// libgovernor regime.
// Set by execution a custom MySQL command from db_governor.
//
// It means the level of Governor intervention into MySQL execution,
// DON'T confuse it with Full/Light Mode.
//
extern volatile int gov_regime;
enum gov_regimes
{
	govregime_OFF = 0,
	govregime_COLLECT_STATS_ONLY = 1,
	govregime_COLLECT_STATS_AND_APPLY_LVE = 2
};

// Set libgovernor regime
void govsql_enable();
void govsql_enable_reconn();
void govsql_enable_lve();
void govsql_enable_reconn_lve();

// Expose libgovernor regime as MySQL variable - to be used in 'SHOW VARIABLES', 'SET', etc.
#define GOV_SQLVARADDR_ENABLE()		(const_cast<int*>(&gov_regime))	// cast away 'volatile'

// Implement custom MySQL commands, changing libgovernor regime and forcing reconnection to db_governor
#define GOV_SQLCMDIMPL_ENABLE()						govsql_enable()
#define GOV_SQLCMDIMPL_ENABLE_AND_RECONNECT()		govsql_enable_reconn()
#define GOV_SQLCMDIMPL_ENABLE_LVE()					govsql_enable_lve()
#define GOV_SQLCMDIMPL_ENABLE_LVE_AND_RECONNECT()	govsql_enable_reconn_lve()


// --------------------------------------------------------------
// LVE entering and exit.
//
extern int (*libgovernor_send_info_begin)(const char*);
extern int (*libgovernor_send_info_end)(const char*);

extern int (*govlve_enter)(const char *user);
extern void (*govlve_exit)();

#define _GOV_PREPARE_FOR_LVE(thd, mSecCtx, mUserName) \
	int govRegime = gov_regime;  /* use the SAME snapshot of this volatile for ALL our checks */ \
	/* Make fast and cheap checks first: */ \
	bool govSendStats = libgovernor_send_info_begin && libgovernor_send_info_end && govRegime != govregime_OFF;  /* both functions loaded from libgovernor, and Governor functionality enabled */ \
	bool govEnterLVE = govlve_enter && govlve_exit && govRegime == govregime_COLLECT_STATS_AND_APPLY_LVE;  /* Governor LVE functionality is properly initialized (mode is 'Light or 'Full', but not 'No'), and it is currently enabled by 'gov_regime' value */ \
	std::string govUserName;  /* to use in ALL our actions */ \
	/* If anything is required, make more checks: */ \
	if (govSendStats || govEnterLVE) \
	{ \
		if ((  /* none of Governor custom SQL commands: */ \
			   thd->lex->sql_command != SQLCOM_ENABLE_GOVERNOR \
			&& thd->lex->sql_command != SQLCOM_ENABLE_RECONN_GOVERNOR \
			&& thd->lex->sql_command != SQLCOM_ENABLE_GOVERNOR_LVE \
			&& thd->lex->sql_command != SQLCOM_ENABLE_RECONN_GOVERNOR_LVE) \
			&& /* and non-empty user name: */ \
			(  thd->mSecCtx \
			&& thd->mSecCtx->mUserName \
			&& thd->mSecCtx->mUserName[0])) \
			govUserName = thd->mSecCtx->mUserName;  /* all is OK, save username and do all planned actions on it */ \
		else \
			govSendStats = govEnterLVE = false;  /* wrong SQL command or empty username, do nothing */ \
	}

#define GOV_SEND_INFO_BEGIN(thd, mSecCtx, mUserName) \
	_GOV_PREPARE_FOR_LVE(thd, mSecCtx, mUserName); \
	if (govSendStats)\
		libgovernor_send_info_begin(govUserName.c_str());

#define GOV_SEND_INFO_END() \
	if (govSendStats) \
		libgovernor_send_info_end(govUserName.c_str());

// These two macros are the ultimate purpose of the patch:
#define GOV_LVE_ENTER() \
	int lveEnterResult = -1; \
	if (govEnterLVE) \
	{ \
		lveEnterResult = govlve_enter(govUserName.c_str()); \
		if (lveEnterResult < 0) /* '<0'  ->  error; '0'  ->  entered;  '>0'  ->  no problem, just didn't need to enter */\
			my_error(ER_GET_ERRNO, MYF(0), errno, "Can't enter into LVE"); \
	}

#define GOV_LVE_EXIT() \
	if (lveEnterResult == 0) \
		govlve_exit();
